/* fithc.c
   The FITH programming language. :-)
   azz@josstix.demon.co.uk

   To do:

   * GCing the storage area. (?)
   * Alloc/free.
   * Variables. Compile on a new word which returns an address malloced.

*/

/* Define DEBUG to include debugging facilities using printf. */
/* #define DEBUG 1 */

/* PROTOTYPES: my least favourite C feature */
void compileto(int hi, int lo);
void compiletoword(int w);
int carryon(char c);
int getnextword();
int interpret(int x);
void tabcompilestring(char *s);
void compilelibrary();
int push(int i);
int pop();
int fpush(int i);
int fpop();
void tabinsert(int value);
void compilestring(char *s);
void cword(char *s, int i);
int *findword();

#include <stdio.h>
#include <string.h>

/* The buffer for words as they are read */
#define MAXWORDLEN 80
char thisword[MAXWORDLEN];

/* Core size in words */
#define CORESIZE 4000

/* Initial pointers */
#define TABINIT 10
#define STOREINIT 1000
#define TRANSINIT 2000
#define STACKINIT 3000
#define FSTACKINIT 3500

/* The core and pointers into it */
int core[CORESIZE];

#define TABPTRADDR 0
#define STOREPTRADDR 1
#define TRANSPTRADDR 2
#define STACKPTRADDR 3
#define FSTACKPTRADDR 4

#define TABPTR core[TABPTRADDR]
#define STOREPTR core[STOREPTRADDR]
#define TRANSPTR core[TRANSPTRADDR]
#define STACKPTR core[STACKPTRADDR]
#define FSTACKPTR core[FSTACKPTRADDR]

/* Various flags */
int eofhit, stringmode, compilemode;
int level, skiplevel;

/* The "high byte" used by the 0x01xx and 0x02xx instructions */
int highbyte;

/* Modes for the interpret function */
#define IMMEDIATE 0
#define FUNCTION 1

void compileto(int hi, int lo) {
  compiletoword(256*hi + lo);
}

void compiletoword(int w) {
  if (compilemode) {
#ifdef DEBUG
    printf("compile %04x: %04x\n",STOREPTR,w);
#endif
    core[STOREPTR++] = w;
  } else {
#ifdef DEBUG
    printf("immediate %04x\n",w);
#endif
    interpret(w);
  }
}

void tabcompilestring(char *s) {
  for(;*s;s++) tabinsert(*s);
  tabinsert(0);
}

void compilestring(char *s) {
  for(;*s;s++) compileto(0,*s);
  compileto(0,0);
}

void tabinsert(int value) {
#ifdef DEBUG
  printf("tabinsert %04x: %04x\n",TABPTR,value);
#endif
  core[TABPTR++] = value;
}

int main(int argc, char **argv) {
  int i,j,*m;

  TABPTR = TABINIT;
  STOREPTR = STOREINIT;
  TRANSPTR = TRANSINIT;

  compilelibrary(); /* This also sets TABPTR to point after the seperator */

  STACKPTR = STACKINIT;
  FSTACKPTR = FSTACKINIT;
  eofhit = stringmode = compilemode = 0;
  level=0;
  skiplevel=-1;

  /* Main loop: read words from the input. */
  while(getnextword()) {

#ifdef DEBUG
    printf("parse: \"%s\"\n",thisword);
#endif

    /* Is it a comment? */
    if (!strcmp(thisword,"(")) {
      while((strcmp(thisword,")")) && (!eofhit)) getnextword();
      goto parsenext;
    }
    
    /* Is it a string? */
    if (thisword[0]=='\"') {
      /* The string is now in thisword+1 */
      compileto(0x04,strlen(thisword+1)+1);
      i = compilemode;
      compilemode = 1;
      j = STOREPTR;
      compilestring(thisword+1);
      compilemode = i;
      compileto(0x01,j/256);
      compileto(0x02,j%256);
      goto parsenext;
    }

    /* Is it a character literal? */
    if (thisword[0]=='\'') {
      compileto(0x03,thisword[1]);
      goto parsenext;
    }

    /* No, it's SUPERMAN! */

    /* Is it a number? */
    if (isdigit(thisword[0])) {
      i=atoi(thisword);
      if (i>255) {
	compileto(0x01,i/256);
	compileto(0x02,i%256);
      } else {
	compileto(0x03,i);
      }
      goto parsenext;
    }

    /* Are we defining a new word? */
    if (!strcmp(thisword,":")) {
      compilemode = 1;
      getnextword(); /* its name is now in thisword */
      m = findword();
      if ((m-core)!=TABPTR) { /* then the word is already defined */
	i = *(m+intlen(m)+1); /* i is now the int it compiles to */
	core[TRANSINIT+i-0x1000] = STOREPTR; /* update its translation */
      } else {
	cword(thisword,0x1000+TRANSPTR-TRANSINIT); /* add to the table */
	core[TRANSPTR++] = STOREPTR;
      }
      goto parsenext;
    }

    /* Should we *stop* compiling? */
    if (!strcmp(thisword,";")) {
      compileto(0,0);
      compilemode = 0;
      goto parsenext;
    }

    /* Check if it's a word */
    m = findword();
    if ((m-core)!=TABPTR) { 
      /* Hell, who needs the IOCCC to write confusing code? */
      compiletoword(*(m+intlen(m)+1));
      goto parsenext;
    }

    printf("unrecognised word: \"%s\"\n",thisword);
    
  parsenext: /* is there a better way of doing this? */
  }

  exit(0);
}

int *findword() {
  int *m;
  
  m = core + TABINIT;
  while (((strintcmp(thisword,m))!=0) && ((m-core)<TABPTR)) {
    while (*m) m++;
    m++; m++;
  }
  return m;
}

int intlen(int *q) {
  int i;

  i=0;
  while (*(q++)) i++;
  return i;
}

int strintcmp(char *s, int *q) {
  char tmps[MAXWORDLEN],*t;

  t=tmps;
  while(*q) *(t++)=*(q++);
  *(t++)='\0';
  
  if (strlen(s) != strlen(tmps)) return 1;
  return strcmp(s,tmps);
}

int carryon(char c) {
  if ((c == EOF) || feof(stdin)) eofhit = 1;

  switch(c) {
  case '\"':
    stringmode = !stringmode;
    /* fall through */
  case ' ':
  case '\t':
  case '\n':
  case '\r':
  case EOF:
    return (!stringmode);
  }
  return 0;
}

int getnextword() {
  int p;
  char c;

  p = 0;
  do {
    while((!carryon(c=getchar())) && (p<MAXWORDLEN) && (!eofhit)) {
      thisword[p++] = c;
    }
  } while (p==0);
  thisword[p] = '\0';
  
  return (!eofhit);
}

/* make these check for stack overflows */

int push(int i) {
  core[++STACKPTR] = i;
  return i;
}

int pop() {
  if (STACKPTR>STACKINIT)
    return core[STACKPTR--];
  else
    printf("Stack empty\n");
  return 0;
}

int fpush(int i) {
  core[++FSTACKPTR] = i;
  return i;
}

int fpop() {
  if (FSTACKPTR>FSTACKINIT)
    return core[FSTACKPTR--];
  else
    printf("Fstack empty\n");
  return 0;
}

int interpret(int x) {
  int i, j, k, mode, pc;

  pc=0;
  mode=IMMEDIATE;

  do { 
    if (mode!=IMMEDIATE) x=core[pc++];

#ifdef DEBUG
    printf("pc:%04x op:%04x level:%d skip:%d sp:%d tos:%04x fsp:%d tof:%04x\n",
	   pc,x,level,skiplevel,STACKPTR,core[STACKPTR],FSTACKPTR,
	   core[FSTACKPTR]);
#endif

    /* Cope with the levels */
    switch(x) {
    case 0x0030: /* if */
    case 0x0034: /* case */
    case 0x0036: /* is */
      level++;
      break;
    case 0x0031: /* endif */
    case 0x0035: /* endcase */
    case 0x0037: /* endis */	
      level--;
      break;
    }
    
    if ((skiplevel == level) || (skiplevel == -1)) {
      if (skiplevel == level) skiplevel = -1;
      switch(x/256) {
      case 0x00:
	switch(x) {
	case 0x00:
	  /* Strictly speaking, you won't hit this in immediate mode. */
	  if (mode!=IMMEDIATE) {
	    pc = fpop();
	    if (!pc) mode=IMMEDIATE;
	  }
	  break;
	case 0x01:
	  push(pop()+pop());
	  break;
	case 0x02:
	  i=pop();
	  push(pop()-i);
	  break;
	case 0x03:
	  push(pop()*pop());
	  break;
	case 0x04:
	  i=pop();
	  push(pop()/i);
	  break;
	case 0x05:
	  i=pop();
	  push(pop()%i);
	  break;
	case 0x06:
	  push(pop()==pop());
	  break;
	case 0x07:
	  i=pop();
	  push(pop()<i);
	  break;
	case 0x08:
	  push(!pop());
	  break;
	case 0x09:
	  i=pop(); /* This is necessary as C won't execute the second pop() */
	  j=pop();
	  push(i&&j);
	  break;
	case 0x0a:
	  i=pop();
	  j=pop();
	  push(i||j);
	  break;
	case 0x0b:
	  push(pop()&pop());
	  break;
	case 0x0c:
	  push(pop()|pop());
	  break;
	case 0x0d:
	  push(~pop());
	  break;
	case 0x0e:
	  push(core[pop()]);
	  break;
	case 0x0f:
	  i=pop();
	  core[i]=pop();
	  break;
	case 0x10:
	  i=pop();
	  j=pop();
	  push(i);
	  push(j);
	  break;
	case 0x11:
	  i=pop();
	  push(i);
	  push(i);
	  break;
	case 0x12: 
	  i=pop(); /* ( k j i -- j i k ) */
	  j=pop();
	  k=pop();
	  push(j);
	  push(i);
	  push(k);
	  break;
	case 0x13:
	  i=pop();
	  push(core[STACKPTR-i]);
	  break;
	case 0x14:
	  STACKPTR--;
	  break;
	case 0x15: /* i */
	  push(core[FSTACKPTR]);
	  break;
	case 0x20:
	  putchar(pop());
	  break;
	case 0x21:
	  push(getchar());
	  break;
	case 0x22: /* FIXME */
	  i=pop();
	  push(STOREPTR);
	  STOREPTR=STOREPTR+i; 
	  break;
	case 0x23: /* WRITEME */
	  break;
	case 0x30: /* if */
	  if (!pop()) skiplevel=level-1;
	  break;
	case 0x31: /* endif -- does nothing */
	  break;
	case 0x32: /* while */
	  fpush(pc);
	  break;
	case 0x33: /* endwhile */
	  i=fpop();
	  if (pop()) {
	    pc=i;
	    fpush(i);
	  }
	  break;
	case 0x34: /* case */
	  fpush(pop());
	  break;
	case 0x35: /* endcase */
	  fpop();
	  break;
	case 0x36: /* is */
	  if (pop()!=core[FSTACKPTR]) skiplevel=level-1;
	  break;
	case 0x37: /* endis */
	  /*	  skiplevel=level-1;*/ /* jump to the endcase */
	  break;
	case 0x38: /* for */
	  i=pop(); /* initial value */
	  j=pop(); /* final value */
	  fpush(pc);
	  fpush(j);
	  fpush(i);
	  break;
	case 0x39: /* endfor */
	  i=fpop();
	  j=fpop();
	  k=fpop();
	  if (i<j) {
	    pc=k;
	    fpush(k);
	    fpush(j);
	    fpush(i+1);
	  }
	  break;
	case 0x50:
	  push(TABPTRADDR);
	  break;
	case 0x51:
	  push(STOREPTRADDR);
	  break;
	case 0x52:
	  push(TRANSPTRADDR);
	  break;
	case 0x53:
	  push(STACKPTRADDR);
	  break;
	case 0x54:
	  push(FSTACKPTRADDR);
	  break;
	case 0x60:
	  push(TABINIT);
	  break;
	case 0x61:
	  push(STOREINIT);
	  break;
	case 0x62:
	  push(TRANSINIT);
	  break;
	case 0x63:
	  push(STACKINIT);
	  break;
	case 0x64:
	  push(FSTACKINIT);
	  break;
	default:
	  puts("Unknown compiled op\n");
	}
	break;
      case 0x01:
	highbyte = x%256;
	break;
      case 0x02:
	push((highbyte*256) + (x%256));
	break;
      case 0x03:
	push(x%256);
	break;
      case 0x04:
	if (mode!=IMMEDIATE) {
	  pc = pc + (x%256);
	} /* In immediate mode, ignore this op. */
	break;
      default:
	if (x>=0x1000) {
	  fpush(pc);
	  pc = core[TRANSINIT+x-0x1000];
	  mode=FUNCTION;
	} 
	break;
      }
    }
  } while (mode!=IMMEDIATE);
}

void cword(char *x, int y) {
  tabcompilestring(x);
  tabinsert(y);
}

void compilelibrary() {
  compilemode = 1;

  /* Output all the builtins */
  cword("+",0x01);
  cword("-",0x02);
  cword("*",0x03);
  cword("/",0x04);
  cword("mod",0x05);
  cword("=",0x06);
  cword("<",0x07);
  cword("not",0x08);
  cword("and",0x09);
  cword("or",0x0a);
  cword("band",0x0b);
  cword("bor",0x0c);
  cword("bnot",0x0d);
  cword("@",0x0e);
  cword("!",0x0f);
  cword("swap",0x10);
  cword("dup",0x11);
  cword("rot",0x12);
  cword("pick",0x13);
  cword("drop",0x14);
  cword("i",0x15);
  cword("putchar",0x20);
  cword("getchar",0x21);
  cword("alloc",0x22);
  cword("if",0x30);
  cword("endif",0x31);
  cword("while",0x32);
  cword("endwhile",0x33);
  cword("case",0x34);
  cword("endcase",0x35);
  cword("is",0x36);
  cword("endis",0x37);
  cword("for",0x38);
  cword("endfor",0x39);
  cword("tabbase",0x40);
  cword("transbase",0x41);
  cword("tabptr",0x50);
  cword("storeptr",0x51);
  cword("transptr",0x52);
  cword("stackptr",0x53);
  cword("fstackptr",0x54);
  cword("tabinit",0x60);
  cword("storeinit",0x61);
  cword("transinit",0x62);
  cword("stackinit",0x63);
  cword("fstackinit",0x64);
  
  compilemode = 0;
}
